/*************************************************************************
HPGComm (Unix Version)
**************************************************************************

Copyright 2003 Joshua King, ported from HPGComm for Windows released by
HP and Jean-Pierre Bergamin

This file is part of HPGComm (Unix Version).

HPGComm is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

HPGComm is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HPGComm; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

**************************************************************************

Copyright (C) January 2001 by HP and Jean-Pierre Bergamin

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
***************************************************************************/


// Image.h: interface for the CImage class.
//
//////////////////////////////////////////////////////////////////////

#include "notwindows.h"

#if !defined(AFX_IMAGE_H__7B01B634_27D4_11D3_81A7_0000B4A265C3__INCLUDED_)
#define AFX_IMAGE_H__7B01B634_27D4_11D3_81A7_0000B4A265C3__INCLUDED_

#ifndef _NOT_WINDOWS_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#endif // _NOT_WINDOWS_

class CImage  
{	
public:
	typedef struct {
		LONG height;	// Height of the image
		LONG width;		// Width of the image
		WORD planes;	// GROB planes
		int colors;	// Number of colors in the image.
		DWORD size;		// Size of the image
		LONG bytesPerLine;	// How many bytes are in one BMP scan line with valid data
		LONG fourBytes;		// How many bytes are in one BMP scan line (DWORD alignement)
	} tagImageInfo; 

	CImage();
	virtual ~CImage();

    int GetBitCount();
	BYTE * GetBits();
#ifndef _NOT_WINDOWS_
	CSize GetSize();
#endif // _NOT_WINDOWS_
	bool isGrayscale();
	bool IsBitSet(BYTE val, int nBit);

#ifndef _NOT_WINDOWS_
	bool Alloc(DWORD size);
    void Free();
    void FillHeaderStruct(BITMAPINFOHEADER &bmiHeader, tagImageInfo * pInfo = 0);
#endif // _NOT_WINDOWS_
	void CreateGrayMap(int nColors, BYTE *& cursor);
#ifndef _NOT_WINDOWS_
	HBITMAP CreateBitmapHandle(BYTE * pIm = 0 );
	HBITMAP GetHandle();
#endif // _NOT_WINDOWS_
	
    bool LoadScreen(unsigned char * screendata);
#ifndef _NOT_WINDOWS_
	BOOL LoadBMP( LPCTSTR sBMPFile);
	BOOL LoadGROB(LPCTSTR sGROBFile, int nPlanes = 0);
	BOOL SaveGROB(CString * pFileName);
#endif // _NOT_WINDOWS_
	BOOL SaveBMP(CString * pFileName);
#ifndef _NOT_WINDOWS_
    BOOL Update(HBITMAP hBM);
    bool Border(bool bBlack, bool bThin, bool bAdd);
	void CopyToClipboard( const CWnd *pWnd);
	void Reset();
	void Magnify(BOOL bMagnify = TRUE);
	void Rotate();
	void InvertBitmap();
	void MirrorBitmap(BOOL bLateral);
	bool ConvertToGray();
	void ConvertToMono();
	void Resize(CSize size);

	HBITMAP origSize;
	bool bBorder;
	bool bBorderThin;
	bool bBorderBlack;
	HBITMAP origBitmap;
	BYTE * phDIB;
#endif // _NOT_WINDOWS_
	
private:
	BYTE TurnByte(BYTE byte);
	void SetBit(int nBit, BYTE &byte, BOOL bFalse);
#ifndef _NOT_WINDOWS_
	HANDLE DDBToDIB( CBitmap& bitmap, DWORD dwCompression);
#endif // _NOT_WINDOWS_
	BYTE TurnNibble(BYTE value);

    tagImageInfo imageInfo;
#ifndef _NOT_WINDOWS_
	HGLOBAL hDIB;
	//BITMAPINFO * bmInfo;
	//BITMAPFILEHEADER bmFileHeader;
	CBitmap bitmap;
#else
	CBMPFile bitmap;
#endif // _NOT_WINDOWS_
	int planes;

protected:
	void ConvertGROBToBMP(BYTE *&bmp, BYTE * grob);
};

#endif // !defined(AFX_IMAGE_H__7B01B634_27D4_11D3_81A7_0000B4A265C3__INCLUDED_)
