/*************************************************************************
HPGComm (Unix Version)
**************************************************************************

Copyright 2003 Joshua King, ported from HPGComm for Windows released by
HP and Jean-Pierre Bergamin

This file is part of HPGComm (Unix Version).

HPGComm is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

HPGComm is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HPGComm; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

***************************************************************************/

#include "bmpfile.h"

// Constructor

CBMPFile::CBMPFile() {
	colourTable = NULL;
	rasterData = NULL;
}

// Destructor

CBMPFile::~CBMPFile() {
	if (colourTable != NULL) {
		delete [] colourTable;
	}
	if (rasterData != NULL) {
		delete [] rasterData;
	}
}

// Member functions

void CBMPFile::longToCharArray(unsigned long in, unsigned char * buf) {
	// This will write the long as 4 bytes Intel-style
	int i;
	
	for (i = 0; i < 4; i++) {
		*(buf++) = (unsigned char) (in % 256);
		in = in >> 8;
	}
}
	
void CBMPFile::shortToCharArray(unsigned short in, unsigned char * buf) {
	// This will write the short as 4 bytes Intel-style
	int i;
	
	for (i = 0; i < 2; i++) {
		*(buf++) = (unsigned char) (in % 256);
		in = in >> 8;
	}
}

unsigned long CBMPFile::charArrayToLong(unsigned char * buf) {
	int i;
	unsigned long out = 0;
	
	for (i = 0; i < 4; i++) {
		out |= (*(buf++) << (8 * i));
	}
	
	return out;
}

unsigned short CBMPFile::charArrayToShort(unsigned char * buf) {
	int i;
	unsigned short out = 0;
	
	for (i = 0; i < 2; i++) {
		out |= (*(buf++) << (8 * i));
	}
	
	return out;
}

void CBMPFile::InitBMP(unsigned long width, unsigned long height, unsigned short colours)
{
	fileHeader.signature[0] = 'B';
	fileHeader.signature[1] = 'M';
	longToCharArray(calcFileSize(width, height, colours), fileHeader.fileSize);
	longToCharArray(0, fileHeader.reserved);
	longToCharArray(calcDataOffset(colours), fileHeader.dataOffset);
	longToCharArray(40, infoHeader.size);
	longToCharArray(width, infoHeader.width);
	longToCharArray(height, infoHeader.height);
	shortToCharArray(1, infoHeader.planes);
	shortToCharArray((unsigned short) (logBase2(colours)), infoHeader.bitCount);
	longToCharArray(0, infoHeader.compression);
	longToCharArray(0, infoHeader.imageSize);
	longToCharArray(2000, infoHeader.xPixelsPerMetre);
	longToCharArray(2000, infoHeader.yPixelsPerMetre);
	longToCharArray(colours, infoHeader.coloursUsed);
	longToCharArray(0, infoHeader.coloursImportant);
	if (colourTable != NULL)
	{
		delete [] colourTable;
		colourTable = NULL;
	}
	colourTable = new struct COLOURTABLEENTRY[colours];
	if (rasterData != NULL)
	{
		delete [] rasterData;
		rasterData = NULL;
	}
	int arrayLength = calcBytesPerLine(width, colours) * height;
	rasterData = new unsigned char[arrayLength];
	int i;
	for (i = 0; i < arrayLength; i++) {
		rasterData[i] = 0;
	}
}

bool CBMPFile::SaveBMP(string filename)
{
	if (colourTable == NULL || rasterData == NULL) {
		return false;
	}

	int fileDescriptor = open(filename.c_str(), O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
	
	if (fileDescriptor != -1)
	{
		write(fileDescriptor, &fileHeader, sizeof(fileHeader));
		write(fileDescriptor, &infoHeader, sizeof(infoHeader));
		int i;
		for (i = 0; i < (1 << charArrayToShort(infoHeader.bitCount)); i++) {
			write(fileDescriptor, &colourTable[i], sizeof(COLOURTABLEENTRY));
		}
		write(fileDescriptor, rasterData, calcBytesPerLine(charArrayToLong(infoHeader.width), 
		      1 << charArrayToShort(infoHeader.bitCount)) * charArrayToLong(infoHeader.height));
		close(fileDescriptor);
		return true;
	}
	else
	{
		return false;
	}
}

unsigned long CBMPFile::calcBytesPerLine(unsigned long width, unsigned short colours)
{
	unsigned long answer = width * logBase2(colours);
	if (answer % 32 != 0) {
		answer += (32 - answer % 32);
	}
	return answer >> 3;
}

unsigned long CBMPFile::calcFileSize(unsigned long width, unsigned long height, unsigned short colours)
{
	return 14 + 40 + colours * 4 + calcBytesPerLine(width, colours) * height;
}

unsigned long CBMPFile::calcDataOffset(unsigned short colours)
{
	return 14 + 40 + colours * 4;
}

int CBMPFile::logBase2(int value)
{
	int log = 0;
	while (value > 1)
	{
		value = value >> 1;
		log++;
	}
	return log;
}

/*
int main(int argc, const char * argv[])
{
	CBMPFile myBMP;
	myBMP.InitBMP();
	myBMP.SaveBMP("/tmp/test2.bmp");
}
*/
