/*************************************************************************
HPGComm (Unix Version)
**************************************************************************

Copyright 2003 Joshua King, ported from HPGComm for Windows released by
HP and Jean-Pierre Bergamin

This file is part of HPGComm (Unix Version).

HPGComm is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

HPGComm is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HPGComm; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

***************************************************************************/

#include "notwindows.h"

#ifndef __BMPFILE_H__
#define __BMPFILE_H__


// From http://www.daubnet.com/formats/BMP.html

struct BITMAPFILEHEADER {
	// BITMAPFILEHEADER
	char signature[2]; // 2 bytes - must be "BM"
	unsigned char fileSize[4]; // 4 bytes - size of file
	unsigned char reserved[4]; // 4 bytes - must be zero
	unsigned char dataOffset[4]; // 4 bytes - offset to raster data
};

struct BITMAPINFOHEADER {	
	// BITMAPINFOHEADER
	unsigned char size[4]; // 4 bytes - size of info header, should be 40
	unsigned char width[4]; // 4 bytes - width of bitmap
	unsigned char height[4]; // 4 bytes - height of bitmap
	unsigned char planes[2]; // 2 bytes - number of planes = 1
	unsigned char bitCount[2]; // 2 bytes - bits per pixel (monochrome)
	unsigned char compression[4]; // 4 bytes - BI_RGB (no compression)
	unsigned char imageSize[4]; // 4 bytes - 0 acceptable if no compression
	unsigned char xPixelsPerMetre[4]; // 4 bytes - 72 dpi = approx. 72 * 12 * 3 ppm
	unsigned char yPixelsPerMetre[4]; // 4 bytes - 72 dpi
	unsigned char coloursUsed[4]; // 4 bytes - 2 colours
	unsigned char coloursImportant[4]; // 4 bytes - 0 = all important
};

struct COLOURTABLEENTRY {
	// COLOUR TABLE ENTRY
	unsigned char red; // 1 byte - red component of black colour
	unsigned char green; // 1 byte - green component of black colour
	unsigned char blue; // 1 byte - blue component of black colour
	unsigned char reserved; // 1 byte - must be 0
};

class CBMPFile 
{
public:
	struct BITMAPFILEHEADER fileHeader;
	struct BITMAPINFOHEADER infoHeader;
	struct COLOURTABLEENTRY * colourTable;
	unsigned char * rasterData;
	
	CBMPFile();
	~CBMPFile();
	
	static void longToCharArray(unsigned long value, unsigned char * buf);
	static void shortToCharArray(unsigned short value, unsigned char * buf);
	static unsigned long charArrayToLong(unsigned char * buf);
	static unsigned short charArrayToShort(unsigned char * buf);

	void InitBMP(unsigned long width = 131,
	             unsigned long height = 64,
	             unsigned short colours = 2);
	
	bool SaveBMP(string filename);
protected:
	static unsigned long calcBytesPerLine(unsigned long width, unsigned short colours);
	static unsigned long calcFileSize(unsigned long width, unsigned long height, unsigned short colours);
	static unsigned long calcDataOffset(unsigned short colours);
	static int logBase2(int value);
};

#endif // __BMPFILE_H__
