/*************************************************************************
HPGComm (Unix Version)
**************************************************************************

Copyright 2003 Joshua King, ported from HPGComm for Windows released by
HP and Jean-Pierre Bergamin

This file is part of HPGComm (Unix Version).

HPGComm is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

HPGComm is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with HPGComm; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
***************************************************************************/

#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include "Image.h"
#include "Kermit.h"
#include "XModem.h"

#define HPGCOMM_VERSION "1.2"

using namespace std;

CProt * myProt;

void closeport(void)
{
    if (myProt != NULL) {
	myProt->ClosePort();
	delete myProt; // this will call the destructor which will close the port
	cout << "Stopped listening to port." << endl;
    }
}

void dosigclose(int signal)
{
	cout << "Closing on signal #" << signal << endl;
	closeport();
	exit(0);
}

// From HPComm38Dlg.cpp
//void CHPComm38Dlg::ReadScreenData( char ScreenType ) // CHANGED
void readScreenData(char screenType)
{
    //unsigned int nBytesToRead = 1080 - 1; // CHANGED
    int nBytesToRead = 1080 - 1;
    
    switch (screenType) {
	case 27: 
	    nBytesToRead = 1080 - 1;
	    break;
	case 'G':
	    nBytesToRead = 64 * 17 * 2;
	    break;
	case 'W':
	    nBytesToRead = 64 * 17;
	    break;
    }
    
    unsigned char * m_pScreenData = new unsigned char [nBytesToRead+1];
    *m_pScreenData = screenType;
    //DWORD dwRead; // CHANGED
    int iRead;
    //if ((dwRead = m_pProt->ReadBufferFromCom((char *)(m_pScreenData + 1), nBytesToRead)) != nBytesToRead) { // CHANGED
    if ((iRead = myProt->ReadBufferFromCom((char *)(m_pScreenData + 1), nBytesToRead)) != nBytesToRead) {
	cout << "SCREEN: Problem reading screen data." << endl;
	//MessageBox("There was a problem collecting the screen image", "Screen image error", MB_OK|MB_ICONEXCLAMATION); // CHANGED
	delete [] m_pScreenData;
	m_pScreenData = 0;
    } else {
	    CImage myImage;
	    if (myImage.LoadScreen(m_pScreenData)) {
		    cout << "SCREEN: Received screen capture. (BMP format)" << endl;
		    char * sNewReqFileName = new char[256];
		    cout << "SCREEN: Enter the name of the file to save into and press ENTER." << endl; 
		    cin.getline(sNewReqFileName, 256);
		    if (strlen(sNewReqFileName) > 0) {
			    if (access(sNewReqFileName, F_OK | W_OK) == 0) {
				    cout << "SCREEN: Warning file exists and will be overwritten!" << endl;
			    } else if (access(sNewReqFileName, F_OK) == 0) {
				    cout << "SCREEN: File exists and cannot be overwritten." << endl;
				    delete [] sNewReqFileName;
				    return;
			    }
			    string sReqFileName(sNewReqFileName);
			    if (myImage.SaveBMP(&sReqFileName)) {
				    cout << "SCREEN: Screen capture saved." << endl;
			    } else {
				    cout << "SCREEN: Problem saving screen capture." << endl;
			    }
		    }
		    delete [] sNewReqFileName;
	    } else {
		    cout << "SCREEN: Couldn't decode screen capture." << endl;
	    }
	    delete [] m_pScreenData;
    }
}

int main(int argc, const char * argv[])
{
	bool badargs = false;
	bool verbose = false;
	int i;
	int portArg = -1;
	
	for (i = 1; i < argc; i++) {
		if (*argv[i] == '-') {
			// Is a command-line option
			if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0) {
				// Display usage
				badargs = true;
				break;
			} else if (strcmp(argv[i], "-v") == 0 || strcmp(argv[i], "--verbose") == 0) {
				// Display verbose progress
				verbose = true;
			} else {
				cout << "Unrecognized option '" << argv[i] << "'" << endl;
				badargs = true;
				break;
			}
		} else if (portArg == -1) {
			portArg = i;
		} else {
			cout << "Too many options: " << argc << endl;
			badargs = true;
			break;
		}
	}
	
    if (argc <= 1 || badargs || portArg < 1) {
	    cout << "HPGComm modified for Unix, version " << HPGCOMM_VERSION << endl;
	    cout << "Copyright 2003 Joshua King," << endl;
	    cout << "Derived from HPGComm for Windows," << endl;
	    cout << "Copyright January 2001 HP and Jean-Pierre Bergamin." << endl;
	    cout << "This program is licensed under the GNU General Public License, version 2." << endl;
	    cout << "For more information, see the COPYING file distributed with this program." << endl;
	    cout << "\nUsage:" << endl;
	    cout << argv[0] << " <path to serial port>" << endl;
	    cout << "Please run this program from the directory containing your calculator files." << endl;
	    cout << "Press Control+C or Control+/ to stop the program." << endl;
	    return badargs ? 2 : 0;
    }
    
    myProt = new CKermit();
	myProt->m_pTransferDlg = verbose;
    string sPath(getcwd(NULL, 0));
    
    try {
		myProt->OpenPort(argv[portArg]);
    }
    catch (CProtException * pEx) {
		// Ensure the dialog is friendlier when we first start.
		if (pEx->error == CProtException::OPENPORT) {
			cout << "Error #1: Could not open serial port." << endl;
			return 1;
		} else {
			cout << "Unknown error occurred." << endl;
			return 127;
		}
    }

    signal(SIGHUP, dosigclose);
    signal(SIGINT, dosigclose);
    signal(SIGQUIT, dosigclose);
    signal(SIGTERM, dosigclose);
    
    myProt->CreateComDaemon(); // not necessary here
    myProt->UnleashDaemon(); // not necessary here
    
    bool bIsXModem = false;
    char cFirstChar;
    char cSecondChar;
    CProt * pOld;
    
    cout << "HPGComm for Unix/Mac OS X, version " << HPGCOMM_VERSION << endl;
    cout << "Run with no parameters to see copyright, licensing and usage information." << endl;
    cout << "Listening on " << argv[portArg] << endl;
    cout << "Press CONTROL+C (^C) to stop listening..." << endl;
    
    do {
	pOld = myProt;
	try {
            // First we have to check if the data at the Com-port is a screenshot
            // of a kermit-packet.
            // If the first char is 27 it's a screenshot for 48 & 38
            // If the second char is 'W' or 'G', it is a screenshot for 49/39/40
            // Otherwise it's a kermit-packet.
            if (myProt->ReadCharFromCom(&cFirstChar) == 0) {
		cout << "TIMEOUT: Please make sure that the calculator is connected." << endl;
		throw new CProtException(CProtException::TIMEOUT);
            }
	    if (cFirstChar == 0x18) {
		// This is the CAN char of the XModem protocol to cancel a transfer
  // Ignore this as the first character...
		cout << "XMODEM: The transfer was cancelled." << endl;
		continue;
	    }
            if (cFirstChar == 0x15 || cFirstChar == 'D') {
		cout << "XMODEM: Starting XModem transaction..." << endl;
		CXModem XModem(myProt->hComPort);
		myProt->m_pTransferDlg = verbose;
		bIsXModem = true;
		XModem.m_sCurPCPath = sPath;
		myProt = &XModem;
		try {
			XModem.Transfer(true);
		} catch (CProtException * pEx2) {
			cout << "XMODEM: Transfer cancelled." << endl;
		}
		myProt = pOld;
		myProt->FlushComPort();
		//myProt->UnleashDaemon();
		bIsXModem = false;
	    }
            else if (cFirstChar == 27) {
                // We have a screen shot waiting to be shown...
                readScreenData(cFirstChar);
            }
            else {
                if (!myProt->ReadCharFromCom(&cSecondChar)) {
                    throw new CProtException(CProtException::TIMEOUT);
                }
		
                if ((cSecondChar == 'W') || (cSecondChar == 'G')) {
		    readScreenData(cSecondChar);
                }
                else {
					//cout << "KERMIT: Starting Kermit transaction..." << endl;
                    myProt->ClientRequest(cFirstChar, cSecondChar, sPath);
				}
            }
        }
	
	catch (CProtException * pEx) {
	    myProt->Reset();
	    
	    continue;
	}
	
	// If the transaction was finished (this means the 38G wrote
 // a directory file to disk and a new file was created on the
 // PC, we reread the content of the PC dir, too.
	if (myProt->m_bTransactionFinished) {
	    myProt->m_bNewFileCreated = false;
	    myProt->m_bTransactionFinished = false;
	}
    } while (true);
    
    delete myProt;
    
    return 0;
}

